#!/usr/bin/env python
"""Simple STUN client for getting external IP address.
Based on stun.py from https://github.com/myers/html5_udp_to_server,
but heavily simplified (dropped Twisted dependency, only supports old RFC 3489)
"""

import os, socket, struct, math

BINDING_REQUEST = 0x0001
BINDING_RESPONSE = 0x0101
MAPPED_ADDRESS = 0x0001
FAMILY_IPV4 = 0x01

STUN_SERVERS = [
    ("stun.l.google.com", 19302),
    ("stun1.l.google.com", 19302),
    ("stun2.l.google.com", 19302),
    ("stun3.l.google.com", 19302),
    ("stun4.l.google.com", 19302),
    ("stun.iptel.org", 3478)
]


class StunClient(object):
    """Simple STUN client for getting external IP address"""
    def __init__(self, timeout=5, attempts=3):
        self._timeout = timeout
        self._attempts = attempts
        self._transaction_id = b""

    def get_ip(self, stun_host="", stun_port=3478, source_address="", source_port=54320):
        """Get external IP address and port"""
        if stun_host:
            stun_servers = [(stun_host, stun_port)]
        else:
            stun_servers = STUN_SERVERS

        # If no stun_host is specified, try more servers from STUN_SERVERS list
        # (max. number of attempted servers is specified by self._attempts)
        ext_address = ""
        ext_port = 0
        attempt = 0
        for stun_host, stun_port in stun_servers:
            attempt += 1
            if attempt > self._attempts:
                break

            try:
                # Create socket, send request and receive response
                sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                sock.settimeout(self._timeout)
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                sock.bind((source_address, source_port))
                sock.sendto(self._generate_request(), (stun_host, stun_port))
                data, addr = sock.recvfrom(2048)
                sock.close()

                # Parse response
                ext_address, ext_port = self._parse_response(data)
                if ext_address:
                    break
            except:
                continue

        if not ext_address:
            raise RuntimeError("Couldn't get external IP address from STUN server!")

        return (ext_address, ext_port)

    def _generate_id(self):
        """Generate random Transaction ID"""
        return os.urandom(16)

    def _generate_request(self):
        """Generate Binding Request"""
        self._transaction_id = self._generate_id()
        request = [struct.pack(">H", BINDING_REQUEST),  # Message Type
                   struct.pack(">H", 0),                # Message Length
                   self._transaction_id]
        return b"".join(request)

    def _parse_response(self, data):
        """Parse server response to get mapped address"""
        packet_type, length = struct.unpack(">2H", data[:4])
        if packet_type != BINDING_RESPONSE:
            raise ValueError("Invalid response type!")
        if data[4:20] != self._transaction_id:
            raise ValueError("Invalid response transaction ID!")

        # Walk through all response attributes to find MAPPED_ADDRESS
        for attr_id, value_length, value, start_offset in self._parse_attributes(data[20:length + 20]):
            if attr_id == MAPPED_ADDRESS:
                ip_address, port = self._parse_mapped_address(value)
                break

        return (ip_address, port)

    def _parse_attributes(self, data):
        """Generator which walks through response attributes"""
        ptr = 0
        while ptr < len(data):
            attr_type, length = struct.unpack(">2H", data[ptr:ptr + 4])
            yield attr_type, length, data[ptr + 4:ptr + 4 + length], ptr + 20
            attr_length = 4 + 4 * int(math.ceil(length / 4.0))
            ptr += attr_length

    def _parse_mapped_address(self, value):
        """Get IP address and port from MAPPED_ADDRESS attribute"""
        family, recv_port = struct.unpack(">xBH", value[:4])
        if family != FAMILY_IPV4:
            raise ValueError("IPv6 is not supported!")
        ip_address = socket.inet_ntoa(value[4:])
        return (ip_address, recv_port)


if __name__ == "__main__":
    stun = StunClient()
    ext_address, ext_port = stun.get_ip()
    print("External IP: {}".format(ext_address))
    print("External port: {}".format(ext_port))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'qopenvpn/qopenvpnlogviewer.ui'
#
# Created by: PyQt5 UI code generator 5.8
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_QOpenVPNLogViewer(object):
    def setupUi(self, QOpenVPNLogViewer):
        QOpenVPNLogViewer.setObjectName("QOpenVPNLogViewer")
        QOpenVPNLogViewer.resize(1000, 560)
        self.gridLayout = QtWidgets.QGridLayout(QOpenVPNLogViewer)
        self.gridLayout.setObjectName("gridLayout")
        self.logViewerEdit = QtWidgets.QPlainTextEdit(QOpenVPNLogViewer)
        self.logViewerEdit.setTextInteractionFlags(QtCore.Qt.TextSelectableByKeyboard|QtCore.Qt.TextSelectableByMouse)
        self.logViewerEdit.setObjectName("logViewerEdit")
        self.gridLayout.addWidget(self.logViewerEdit, 0, 0, 1, 3)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label = QtWidgets.QLabel(QOpenVPNLogViewer)
        self.label.setObjectName("label")
        self.horizontalLayout.addWidget(self.label)
        self.ipAddressEdit = QtWidgets.QLineEdit(QOpenVPNLogViewer)
        self.ipAddressEdit.setReadOnly(True)
        self.ipAddressEdit.setObjectName("ipAddressEdit")
        self.horizontalLayout.addWidget(self.ipAddressEdit)
        self.gridLayout.addLayout(self.horizontalLayout, 1, 0, 1, 2)
        self.refreshButton = QtWidgets.QPushButton(QOpenVPNLogViewer)
        self.refreshButton.setObjectName("refreshButton")
        self.gridLayout.addWidget(self.refreshButton, 2, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(453, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 2, 1, 1, 1)
        self.buttonBox = QtWidgets.QDialogButtonBox(QOpenVPNLogViewer)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Close)
        self.buttonBox.setObjectName("buttonBox")
        self.gridLayout.addWidget(self.buttonBox, 2, 2, 1, 1)

        self.retranslateUi(QOpenVPNLogViewer)
        self.buttonBox.accepted.connect(QOpenVPNLogViewer.accept)
        self.buttonBox.rejected.connect(QOpenVPNLogViewer.reject)
        QtCore.QMetaObject.connectSlotsByName(QOpenVPNLogViewer)

    def retranslateUi(self, QOpenVPNLogViewer):
        _translate = QtCore.QCoreApplication.translate
        QOpenVPNLogViewer.setWindowTitle(_translate("QOpenVPNLogViewer", "QOpenVPN Log Viewer"))
        self.label.setText(_translate("QOpenVPNLogViewer", "IP address:"))
        self.refreshButton.setText(_translate("QOpenVPNLogViewer", "Refresh"))




#!/usr/bin/env python

import sys, os, subprocess, socket, glob, signal
from PyQt5 import QtCore, QtGui, QtWidgets

from qopenvpn import stun
from qopenvpn.ui_qopenvpnsettings import Ui_QOpenVPNSettings
from qopenvpn.ui_qopenvpnlogviewer import Ui_QOpenVPNLogViewer


# Allow CTRL+C and/or SIGTERM to kill us (PyQt blocks it otherwise)
signal.signal(signal.SIGINT, signal.SIG_DFL)
signal.signal(signal.SIGTERM, signal.SIG_DFL)


class QOpenVPNSettings(QtWidgets.QDialog, Ui_QOpenVPNSettings):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setupUi(self)

        settings = QtCore.QSettings()
        self.sudoCommandEdit.setText(settings.value("sudo_command") or "kdesu")
        self.sudoCheckBox.setChecked(settings.value("use_sudo", False, type=bool))
        self.warningCheckBox.setChecked(settings.value("show_warning", False, type=bool))

        # Checks for the new location of OpenVPN configuration files introduced in OpenVPN 2.4
        # See https://github.com/OpenVPN/openvpn/blob/master/Changes.rst#user-visible-changes
        # "The configuration files are picked up from the /etc/openvpn/server/ and
        # /etc/openvpn/client/ directories (depending on unit file)."
        # Remove this unaesthetic version check when openvpn 2.4 is widely accepcted
        try:
            output = subprocess.check_output(["/usr/bin/env", "openvpn", "--version"])
        except subprocess.CalledProcessError as e:
            output = e.output
        except OSError:
            print("OpenVPN executable not found!", file=sys.stderr)
            output = ""

        # Take second tuple of version output (i.e. `2.4.0`)
        # and extract its major and minor components (i.e. 2 and 4)
        version_string = output.decode("utf8").split()[1] if output else ""
        version_components = version_string.split(".")
        if len(version_components) >= 2:
            major, minor = version_components[0:2]
            major = int(major)
            minor = int(minor)
        else:
            print("Couldn't determine the installed OpenVPN version, assuming v0.0", file=sys.stderr)
            major = minor = 0

        # Matches version 2.4.x or greater
        if major >= 2 and minor >= 4:
            settings.setValue("config_location", "/etc/openvpn/client/*.conf")
            settings.setValue("service_name", "openvpn-client")
        else:
            settings.setValue("config_location", "/etc/openvpn/*.conf")
            settings.setValue("service_name", "openvpn")

        # Fill VPN combo box with .conf files from /etc/openvpn{,/client}
        for f in sorted(glob.glob(settings.value("config_location"))):
            vpn_name = os.path.splitext(os.path.basename(f))[0]
            self.vpnNameComboBox.addItem(vpn_name)

        i = self.vpnNameComboBox.findText(settings.value("vpn_name"))
        if i > -1:
            self.vpnNameComboBox.setCurrentIndex(i)

    def accept(self):
        settings = QtCore.QSettings()
        settings.setValue("sudo_command", self.sudoCommandEdit.text())
        settings.setValue("use_sudo", self.sudoCheckBox.isChecked())
        settings.setValue("show_warning", self.warningCheckBox.isChecked())
        settings.setValue("vpn_name", self.vpnNameComboBox.currentText())
        QtWidgets.QDialog.accept(self)


class QOpenVPNLogViewer(QtWidgets.QDialog, Ui_QOpenVPNLogViewer):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setupUi(self)
        self.refreshButton.clicked.connect(self.refresh)
        self.refresh()

    def journalctl(self, disable_sudo=False):
        """Run journalctl command and get OpenVPN logs"""
        settings = QtCore.QSettings()
        cmdline = []
        if not disable_sudo and settings.value("use_sudo", type=bool):
            cmdline.append(settings.value("sudo_command") or "sudo")
        cmdline.extend([
            "journalctl", "-b", "-u",
            "{}@{}".format(settings.value("service_name"), settings.value("vpn_name"))
        ])
        try:
            output = subprocess.check_output(cmdline)
        except subprocess.CalledProcessError as e:
            output = e.output
        return output

    def getip(self):
        """Get external IP address and hostname"""
        try:
            stunclient = stun.StunClient()
            ip, port = stunclient.get_ip()
        except:
            ip = ""

        try:
            hostname = socket.gethostbyaddr(ip)[0]
        except:
            hostname = ""

        return (ip, hostname)

    def refresh(self):
        """Refresh logs"""
        self.logViewerEdit.setPlainText(self.journalctl(disable_sudo=True).decode("utf8"))
        QtCore.QTimer.singleShot(0, self.refresh_timeout)

    def refresh_timeout(self):
        """Move scrollbar to bottom and refresh IP address
        (must be called by single shot timer or else scrollbar sometimes doesn't move)"""
        self.logViewerEdit.verticalScrollBar().setValue(self.logViewerEdit.verticalScrollBar().maximum())

        ip = self.getip()
        self.ipAddressEdit.setText("{} ({})".format(ip[0], ip[1]) if ip[1] else ip[0])


class QOpenVPNWidget(QtWidgets.QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.vpn_enabled = False

        self.create_actions()
        self.create_menu()
        self.create_icon()
        self.update_status()

        # Update status every 10 seconds
        self.timer = QtCore.QTimer(self)
        self.timer.timeout.connect(self.update_status)
        self.timer.start(5000)

        # Setup system tray icon doubleclick timer
        self.icon_doubleclick_timer = QtCore.QTimer(self)
        self.icon_doubleclick_timer.setSingleShot(True)
        self.icon_doubleclick_timer.timeout.connect(self.icon_doubleclick_timeout)

    def create_actions(self):
        """Create actions and connect relevant signals"""
        self.startAction = QtWidgets.QAction(self.tr("&Start"), self)
        self.startAction.triggered.connect(self.vpn_start)
        self.stopAction = QtWidgets.QAction(self.tr("S&top"), self)
        self.stopAction.triggered.connect(self.vpn_stop)
        self.settingsAction = QtWidgets.QAction(self.tr("S&ettings ..."), self)
        self.settingsAction.triggered.connect(self.settings)
        self.logsAction = QtWidgets.QAction(self.tr("Show &logs ..."), self)
        self.logsAction.triggered.connect(self.logs)
        self.quitAction = QtWidgets.QAction(self.tr("&Quit"), self)
        self.quitAction.triggered.connect(self.quit)

    def create_menu(self):
        """Create menu and add items to it"""
        self.trayIconMenu = QtWidgets.QMenu(self)
        self.trayIconMenu.addAction(self.startAction)
        self.trayIconMenu.addAction(self.stopAction)
        self.trayIconMenu.addSeparator()
        self.trayIconMenu.addAction(self.settingsAction)
        self.trayIconMenu.addAction(self.logsAction)
        self.trayIconMenu.addSeparator()
        self.trayIconMenu.addAction(self.quitAction)

    def create_icon(self):
        """Create system tray icon"""
        self.iconActive = QtGui.QIcon("{}/openvpn.svg".format(os.path.dirname(os.path.abspath(__file__))))
        self.iconDisabled = QtGui.QIcon("{}/openvpn_disabled.svg".format(os.path.dirname(os.path.abspath(__file__))))

        # Workaround for Plasma 5 not showing SVG icons
        self.iconActive = QtGui.QIcon(self.iconActive.pixmap(128, 128))
        self.iconDisabled = QtGui.QIcon(self.iconDisabled.pixmap(128, 128))

        self.trayIcon = QtWidgets.QSystemTrayIcon(self)
        self.trayIcon.activated.connect(self.icon_activated)
        self.trayIcon.setContextMenu(self.trayIconMenu)
        self.trayIcon.setIcon(self.iconDisabled)
        self.trayIcon.setToolTip("QOpenVPN")
        self.trayIcon.show()

    def update_status(self, disable_warning=False):
        """Update GUI according to OpenVPN status"""
        settings = QtCore.QSettings()
        vpn_status = self.vpn_status()
        if vpn_status:
            self.trayIcon.setIcon(self.iconActive)
            self.startAction.setEnabled(False)
            self.stopAction.setEnabled(True)
            self.vpn_enabled = True
        else:
            self.trayIcon.setIcon(self.iconDisabled)
            self.startAction.setEnabled(True)
            self.stopAction.setEnabled(False)

            if not disable_warning and settings.value("show_warning", type=bool) and self.vpn_enabled:
                QtWidgets.QMessageBox.warning(self, self.tr("QOpenVPN - Warning"),
                                              self.tr("You have been disconnected from VPN!"))
            self.vpn_enabled = False

    def systemctl(self, command, disable_sudo=False, quiet=False):
        """Run systemctl command"""
        settings = QtCore.QSettings()
        cmdline = []
        if not disable_sudo and settings.value("use_sudo", type=bool):
            cmdline.append(settings.value("sudo_command") or "sudo")
        cmdline.extend([
            "systemctl", command,
            "{}@{}".format(settings.value("service_name"), settings.value("vpn_name"))
        ])
        stdout = stderr = subprocess.DEVNULL if quiet else None
        return subprocess.call(cmdline, stdout=stdout, stderr=stderr)

    def vpn_start(self):
        """Start OpenVPN service"""
        retcode = self.systemctl("start")
        if retcode == 0:
            self.update_status()

    def vpn_stop(self):
        """Stop OpenVPN service"""
        retcode = self.systemctl("stop")
        if retcode == 0:
            self.update_status(disable_warning=True)

    def vpn_status(self):
        """Check if OpenVPN service is running"""
        retcode = self.systemctl("is-active", disable_sudo=True, quiet=True)
        return True if retcode == 0 else False

    def settings(self):
        """Show settings dialog"""
        dialog = QOpenVPNSettings(self)
        if dialog.exec_() and self.vpn_enabled:
            self.vpn_stop()
            self.vpn_start()

    def logs(self):
        """Show log viewer dialog"""
        dialog = QOpenVPNLogViewer(self)
        dialog.exec_()

    def icon_activated(self, reason):
        """Start or stop OpenVPN by double-click on tray icon"""
        if reason == QtWidgets.QSystemTrayIcon.Trigger or reason == QtWidgets.QSystemTrayIcon.DoubleClick:
            if self.icon_doubleclick_timer.isActive():
                self.icon_doubleclick_timer.stop()
                if self.vpn_enabled:
                    self.vpn_stop()
                else:
                    self.vpn_start()
            else:
                self.icon_doubleclick_timer.start(QtWidgets.QApplication.doubleClickInterval())

    def icon_doubleclick_timeout(self):
        """Action performed after single-click on tray icon"""
        pass

    def quit(self):
        """Quit QOpenVPN GUI (and ask before quitting if OpenVPN is still running)"""
        if self.vpn_enabled:
            reply = QtWidgets.QMessageBox.question(
                self, self.tr("QOpenVPN - Quit"),
                self.tr("You are still connected to VPN! Do you really want to quit "
                        "QOpenVPN GUI (OpenVPN service will keep running in background)?"),
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No, QtWidgets.QMessageBox.No
            )
            if reply == QtWidgets.QMessageBox.Yes:
                QtWidgets.QApplication.quit()
        else:
            QtWidgets.QApplication.quit()


def main():
    app = QtWidgets.QApplication(sys.argv)
    app.setOrganizationName("QOpenVPN")
    app.setOrganizationDomain("qopenvpn.eutopia.cz")
    app.setApplicationName("QOpenVPN")
    app.setQuitOnLastWindowClosed(False)
    window = QOpenVPNWidget()
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'qopenvpn/qopenvpnsettings.ui'
#
# Created by: PyQt5 UI code generator 5.8
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_QOpenVPNSettings(object):
    def setupUi(self, QOpenVPNSettings):
        QOpenVPNSettings.setObjectName("QOpenVPNSettings")
        QOpenVPNSettings.resize(300, 200)
        self.gridLayout = QtWidgets.QGridLayout(QOpenVPNSettings)
        self.gridLayout.setObjectName("gridLayout")
        self.label = QtWidgets.QLabel(QOpenVPNSettings)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)
        self.vpnNameComboBox = QtWidgets.QComboBox(QOpenVPNSettings)
        self.vpnNameComboBox.setObjectName("vpnNameComboBox")
        self.gridLayout.addWidget(self.vpnNameComboBox, 0, 1, 1, 2)
        self.warningCheckBox = QtWidgets.QCheckBox(QOpenVPNSettings)
        self.warningCheckBox.setObjectName("warningCheckBox")
        self.gridLayout.addWidget(self.warningCheckBox, 1, 0, 1, 3)
        self.line_2 = QtWidgets.QFrame(QOpenVPNSettings)
        self.line_2.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.gridLayout.addWidget(self.line_2, 2, 0, 1, 3)
        self.sudoCheckBox = QtWidgets.QCheckBox(QOpenVPNSettings)
        self.sudoCheckBox.setObjectName("sudoCheckBox")
        self.gridLayout.addWidget(self.sudoCheckBox, 3, 0, 1, 2)
        self.label_2 = QtWidgets.QLabel(QOpenVPNSettings)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 4, 0, 1, 2)
        self.sudoCommandEdit = QtWidgets.QLineEdit(QOpenVPNSettings)
        self.sudoCommandEdit.setObjectName("sudoCommandEdit")
        self.gridLayout.addWidget(self.sudoCommandEdit, 4, 2, 1, 1)
        self.buttonBox = QtWidgets.QDialogButtonBox(QOpenVPNSettings)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.gridLayout.addWidget(self.buttonBox, 5, 0, 1, 3)

        self.retranslateUi(QOpenVPNSettings)
        self.buttonBox.accepted.connect(QOpenVPNSettings.accept)
        self.buttonBox.rejected.connect(QOpenVPNSettings.reject)
        QtCore.QMetaObject.connectSlotsByName(QOpenVPNSettings)
        QOpenVPNSettings.setTabOrder(self.vpnNameComboBox, self.warningCheckBox)
        QOpenVPNSettings.setTabOrder(self.warningCheckBox, self.sudoCheckBox)
        QOpenVPNSettings.setTabOrder(self.sudoCheckBox, self.sudoCommandEdit)
        QOpenVPNSettings.setTabOrder(self.sudoCommandEdit, self.buttonBox)

    def retranslateUi(self, QOpenVPNSettings):
        _translate = QtCore.QCoreApplication.translate
        QOpenVPNSettings.setWindowTitle(_translate("QOpenVPNSettings", "QOpenVPN Settings"))
        self.label.setText(_translate("QOpenVPNSettings", "VPN name:"))
        self.warningCheckBox.setText(_translate("QOpenVPNSettings", "Show warning when disconnected"))
        self.sudoCheckBox.setText(_translate("QOpenVPNSettings", "Use sudo"))
        self.label_2.setText(_translate("QOpenVPNSettings", "Sudo command:"))

